from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.chart import ActiveUsersChart, DriveChart

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class ChartActions(AbstractAction):
    def __init__(self, session: HTTPClient, client: ClientManager):
        self.__session = session
        self.__client = client

    async def get_active_user(
        self, span: str = 'day', limit: int = 30, offset: int = 0
    ) -> ActiveUsersChart:
        data = {'span': span, 'limit': limit, 'offset': offset}
        data = await self.__session.request(
            Route('POST', '/api/charts/active-users'), json=data, auth=True, lower=True,
        )
        return ActiveUsersChart(data)

    async def get_drive(self, span: str = 'day', limit: int = 30, offset: int = 0) -> DriveChart:
        data = {'span': span, 'limit': limit, 'offset': offset}
        data = await self.__session.request(
            Route('POST', '/api/charts/drive'), json=data, auth=True, lower=True,
        )
        return DriveChart(data)



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.models.emoji import CustomEmoji
from mipac.types.ads import IAds
from mipac.types.meta import ICPU, ILiteMeta, IMetaAnnouncement, IMetaCommon

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class CPU:
    def __init__(self, cpu: ICPU) -> None:
        self.__cpu: ICPU = cpu

    @property
    def cores(self) -> int:
        return self.__cpu['cores']

    @property
    def model(self) -> str:
        return self.__cpu['model']


class MetaCommon:
    def __init__(self, meta_common: IMetaCommon, *, client: ClientManager) -> None:
        self.__meta_common: IMetaCommon = meta_common
        self.__client = client

    @property
    def cache_remote_files(self) -> bool:
        return self.__meta_common['cache_remote_files']

    @property
    def enable_hcaptch(self) -> bool:
        return self.__meta_common['enable_hcaptch']

    @property
    def hcaptcha_site_key(self) -> str | None:
        return self.__meta_common['hcaptcha_site_key']

    @property
    def enable_recaptcha(self) -> bool:
        return self.__meta_common['enable_recaptcha']

    @property
    def recaptcha_site_key(self) -> str:
        return self.__meta_common['recaptcha_site_key']

    @property
    def sw_publickey(self) -> str | None:
        return self.__meta_common['sw_publickey']

    @property
    def banner_url(self) -> str | None:
        return self.__meta_common['banner_url']

    @property
    def error_image_url(self) -> str | None:
        return self.__meta_common['error_image_url']

    @property
    def icon_url(self) -> str | None:
        return self.__meta_common['icon_url']

    @property
    def max_note_text_length(self) -> int:
        return self.__meta_common['max_note_text_length']

    @property
    def enable_email(self) -> bool:
        return self.__meta_common['enable_email']

    @property
    def enable_twitter_integration(self) -> bool:
        return self.__meta_common['enable_twitter_integration']

    @property
    def enable_github_integration(self) -> bool:
        return self.__meta_common['enable_github_integration']

    @property
    def enable_discord_integration(self) -> bool:
        return self.__meta_common['enable_discord_integration']

    @property
    def enable_service_worker(self) -> bool:
        return self.__meta_common['enable_service_worker']

    @property
    def proxy_account_name(self) -> str | None:
        return self.__meta_common['proxy_account_name']

    @property
    def use_star_for_reaction_fallback(self) -> bool:
        return self.__meta_common.get('use_star_for_reaction_fallback', False)

    @property
    def object_storage_s3_force_path_style(self) -> bool:
        """
        objectStorageでs3ForcePathStyleを使用するかどうか
        注意: v11の管理者のみ取得できます

        Returns
        -------
        bool
            有効かどうか
        """
        return self.__meta_common.get('object_storage_s3_force_path_style', False)

    # v12 only

    @property
    def ads(self) -> list[IAds] | None:  # TODO: モデルに
        return self.__meta_common.get('ads')

    @property
    def translator_available(self) -> bool:
        return self.__meta_common.get('translator_available', False)

    @property
    def email_required_for_signup(self) -> bool:
        return self.__meta_common.get('email_required_for_signup', False)

    @property
    def mascot_image_url(self) -> str | None:
        return self.__meta_common.get('mascot_image_url')

    # v12とv11の共通情報

    @property
    def emojis(self) -> list[CustomEmoji]:
        return (
            [CustomEmoji(i, client=self.__client) for i in self.__meta_common['emojis']]
            if 'emojis' in self.__meta_common
            else []
        )


class LiteMeta(MetaCommon):
    def __init__(self, meta: ILiteMeta, *, client: ClientManager) -> None:
        super().__init__(meta, client=client)

        self.__lite_meta: ILiteMeta = meta

    @property
    def description(self) -> str | None:
        return self.__lite_meta['description']

    @property
    def disable_registration(self) -> bool:
        return self.__lite_meta['disable_registration']

    @property
    def feedback_url(self) -> str:
        return self.__lite_meta['feedback_url']

    @property
    def maintainer_email(self) -> str | None:
        return self.__lite_meta['maintainer_email']

    @property
    def maintainer_name(self) -> str | None:
        return self.__lite_meta['maintainer_name']

    @property
    def name(self) -> str | None:
        return self.__lite_meta['name']

    @property
    def repository_url(self) -> str:
        return self.__lite_meta['repository_url']

    @property
    def tos_url(self) -> str | None:
        return self.__lite_meta['tos_url']

    @property
    def uri(self) -> str:
        return self.__lite_meta['uri']

    @property
    def version(self) -> str:
        return self.__lite_meta['version']

    # v12のMeta

    @property
    def background_image_url(self) -> str | None:
        return self.__lite_meta.get('background_image_url')

    @property
    def default_dark_theme(self) -> str | None:
        return self.__lite_meta.get('default_dark_theme')

    @property
    def default_light_theme(self) -> str | None:
        return self.__lite_meta.get('default_light_theme')

    @property
    def logo_image_url(self) -> str | None:
        return self.__lite_meta.get('logo_image_url')

    @property
    def theme_color(self) -> str | None:
        return self.__lite_meta.get('theme_color')

    # v11のMeta

    @property
    def announcements(self) -> list[IMetaAnnouncement] | None:  # TODO: 型確認
        return self.__lite_meta.get('announcements', None)

    @property
    def cpu(self) -> CPU | None:
        return CPU(self.__lite_meta['cpu']) if 'cpu' in self.__lite_meta else None

    @property
    def disable_local_timeline(self) -> bool:
        return self.__lite_meta.get('disable_local_timeline', False)

    @property
    def disable_global_timeline(self) -> bool:
        return self.__lite_meta.get('disable_global_timeline', False)

    @property
    def enable_emoji_reaction(self) -> bool:
        return self.__lite_meta.get('enable_emoji_reaction', False)

    @property
    def machine(self) -> str | None:
        return self.__lite_meta.get('machine')

    @property
    def node(self) -> str | None:
        return self.__lite_meta.get('node')

    @property
    def os(self) -> str | None:
        return self.__lite_meta.get('os')

    @property
    def proxy_account(self) -> str | None:
        return self.__lite_meta.get('proxy_account')

    @property
    def proxy_remote_files(self) -> bool:
        return self.__lite_meta.get('proxy_remote_files', False)

    @property
    def psql(self) -> str | None:
        return self.__lite_meta.get('psql')

    @property
    def redis(self) -> str | None:
        return self.__lite_meta.get('redis')

    @property
    def secure(self) -> bool:
        return self.__lite_meta.get('secure', False)

    @property
    def summaly_proxy(self) -> str | None:
        return self.__lite_meta.get('summaly_proxy')

    @property
    def tos_text_url(self) -> str | None:
        return self.__lite_meta.get('tos_text_url')

    @property
    def turnstile_secret_key(self) -> str | None:
        return self.__lite_meta.get('turnstile_secret_key')



from __future__ import annotations

from mipac.types.instance import IInstanceLite


class LiteInstance:
    def __init__(self, instance: IInstanceLite) -> None:
        self.__instance: IInstanceLite = instance

    @property
    def name(self) -> str:
        return self.__instance['name']

    def software_name(self) -> str:
        return self.__instance['software_name']

    def software_version(self) -> str:
        return self.__instance['software_version']

    def icon_url(self) -> str:
        return self.__instance['icon_url']

    def favicon_url(self) -> str:
        return self.__instance['favicon_url']

    def theme_color(self) -> str:
        return self.__instance['theme_color']



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator

from mipac.abstract.action import AbstractAction
from mipac.errors.base import NotExistRequiredData, ParameterError
from mipac.http import Route
from mipac.models.emoji import CustomEmoji
from mipac.types.emoji import ICustomEmoji
from mipac.util import check_multi_arg

if TYPE_CHECKING:
    from mipac.http import HTTPClient
    from mipac.manager.client import ClientManager


class AdminEmojiActions(AbstractAction):
    def __init__(self, emoji_id: None | str = None, *, session: HTTPClient, client: ClientManager):
        self.__emoji_id = emoji_id
        self.__session = session
        self.__client = client

    async def add(
        self,
        file_id: str | None = None,
        *,
        name: str | None = None,
        url: str | None = None,
        category: str | None = None,
        aliases: list[str] | None = None
    ) -> bool:
        """絵文字を追加します

        Parameters
        ----------
        file_id : str | None, optional
            追加する絵文字のファイルId, by default None
        name : str | None, optional
            絵文字名, by default None
        url : str | None, optional
            絵文字があるUrl, by default None
        category : str | None, optional
            絵文字のカテゴリー, by default None
        aliases : list[str] | None, optional
            絵文字のエイリアス, by default None

        Returns
        -------
        bool
            成功したかどうか

        Raises
        ------
        NotExistRequiredData
            必要なデータが不足している
        """

        if self.__client._config.use_version >= 12:
            data = {'fileId': file_id}
        else:
            data = {
                'name': name,
                'url': url,
                'category': category,
                'aliases': aliases,
            }

        if not check_multi_arg(file_id, url):
            raise NotExistRequiredData('required a file_id or url')
        return bool(
            await self.__session.request(
                Route('POST', '/api/admin/emoji/add'), json=data, lower=True, auth=True,
            )
        )

    async def gets(
        self,
        query: str | None = None,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        *,
        all: bool = True
    ) -> AsyncGenerator[CustomEmoji, None]:
        if limit > 100:
            raise ParameterError('limitは100以下である必要があります')
        if all:
            limit = 100

        async def request(body) -> list[CustomEmoji]:
            res: list[ICustomEmoji] = await self.__session.request(
                Route('POST', '/api/admin/emoji/list'), auth=True, json=body,
            )
            return [CustomEmoji(emoji, client=self.__client) for emoji in res]

        body = {
            'query': query,
            'limit': limit,
            'sinceId': since_id,
            'untilId': until_id,
        }
        first_req = await request(body)

        for emoji in first_req:
            yield emoji
        if all and len(first_req) == 100:
            body['untilId'] = first_req[-1].id
            while True:
                res = await request(body)
                if len(res) <= 100:
                    for emoji in res:
                        yield emoji
                if len(res) < 100:
                    break

    async def gets_remote(
        self,
        query: str | None = None,
        host: str | None = None,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        *,
        all: bool = True
    ) -> AsyncGenerator[CustomEmoji, None]:
        if limit > 100:
            raise ParameterError('limitは100以下である必要があります')
        if all:
            limit = 100

        async def request(body) -> list[CustomEmoji]:
            res: list[ICustomEmoji] = await self.__session.request(
                Route('POST', '/api/admin/emoji/list-remote'), auth=True, json=body,
            )
            return [CustomEmoji(emoji, client=self.__client) for emoji in res]

        body = {
            'query': query,
            'host': host,
            'limit': limit,
            'sinceId': since_id,
            'untilId': until_id,
        }
        first_req = await request(body)

        for note in first_req:
            yield note

        if all and len(first_req) == 100:
            body['untilId'] = first_req[-1].id
            while True:
                res = await request(body)
                if len(res) <= 100:
                    for note in res:
                        yield note
                if len(res) == 0:
                    break
                body['untilId'] = res[-1].id

    async def remove(self, emoji_id: str | None = None) -> bool:
        """指定したIdの絵文字を削除します

        Parameters
        ----------
        emoji_id : str | None, optional
            削除する絵文字のId, by default None

        Returns
        -------
        bool
            成功したかどうか

        Raises
        ------
        NotExistRequiredData
            Idが不足している
        """

        emoji_id = emoji_id or self.__emoji_id

        if emoji_id is None:
            raise NotExistRequiredData('idが不足しています')

        endpoint = (
            '/api/admin/emoji/delete'
            if self.__client._config.use_version >= 12
            else '/api/admin/emoji/remove'
        )

        return bool(
            await self.__session.request(
                Route('POST', endpoint), auth=True, json={'id': emoji_id}, lower=True,
            )
        )



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, Literal

from mipac.abstract.action import AbstractAction
from mipac.errors.base import NotSupportVersion, NotSupportVersionText, ParameterError
from mipac.http import HTTPClient, Route
from mipac.models.ad import Ad
from mipac.types.ads import IAd

if TYPE_CHECKING:
    from mipac.client import ClientManager


class AdminAdvertisingModelActions(AbstractAction):
    def __init__(self, ad_id: str | None = None, *, session: HTTPClient, client: ClientManager):
        self._ad_id: str | None = ad_id
        self._session: HTTPClient = session
        self._client: ClientManager = client

    async def update(
        self,
        url: str,
        place: Literal['square', 'horizontal', 'horizontal-big'],
        priority: Literal['high', 'middle', 'low'],
        ratio: int,
        image_url: str,
        id: str | None = None,
        starts_at: int = 0,
        expires_at: int = 0,
        memo: str | None = None,
    ) -> bool:
        ad_id = self._ad_id or id
        if ad_id is None:
            raise ParameterError('ad_idは必須です')
        data = {
            'id': ad_id,
            'url': url,
            'memo': memo or '',
            'place': place,
            'priority': priority,
            'ratio': ratio,
            'startsAt': starts_at,
            'expiresAt': expires_at,
            'imageUrl': image_url,
        }
        res: bool = await self._session.request(
            Route('POST', '/api/admin/ad/update'), json=data, auth=True, lower=True
        )
        return res

    async def delete(self, id: str | None = None) -> bool:
        if self._client._config.use_version < 12:
            raise NotSupportVersion(NotSupportVersionText)

        ad_id = self._ad_id or id
        res: bool = await self._session.request(
            Route('POST', '/api/admin/ad/delete'), json={'id': ad_id}, auth=True, lower=True
        )
        return res


class AdminAdvertisingActions(AdminAdvertisingModelActions):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        super().__init__(session=session, client=client)

    async def create(
        self,
        url: str,
        place: Literal['square', 'horizontal', 'horizontal-big'],
        priority: Literal['high', 'middle', 'low'],
        ratio: int,
        image_url: str,
        starts_at: int = 0,
        expires_at: int = 0,
        memo: str | None = None,
    ) -> bool:
        if self._client._config.use_version < 12:
            raise NotSupportVersion(NotSupportVersionText)

        data = {
            'url': url,
            'memo': memo or '',
            'place': place,
            'priority': priority,
            'ratio': ratio,
            'startsAt': starts_at,
            'expiresAt': expires_at,
            'imageUrl': image_url,
        }
        res: bool = await self._session.request(
            Route('POST', '/api/admin/ad/create'), json=data, auth=True, lower=True
        )
        return res

    async def get_list(
        self, limit: int = 10, since_id: str | None = None, until_id: str | None = None
    ) -> AsyncGenerator[Ad, None]:
        if self._client._config.use_version < 12:
            raise NotSupportVersion(NotSupportVersionText)

        async def request(body) -> list[Ad]:
            res: list[IAd] = await self._session.request(
                Route('POST', '/api/admin/ad/list'), lower=True, auth=True, json=body,
            )
            return [Ad(ad, client=self._client) for ad in res]

        data = {'limit': limit, 'sinceId': since_id, 'untilId': until_id}
        if all:
            data['limit'] = 100
        first_req = await request(data)
        for ad in first_req:
            yield ad

        if all and len(first_req) == 100:
            data['untilId'] = first_req[-1].id
            while True:
                res = await request(data)
                if len(res) <= 100:
                    for ad in res:
                        yield ad
                if len(res) == 0:
                    break
                data['untilId'] = res[-1].id



from mipac.types.chart import IActiveUsersChart, IDriveChart, IDriveLocalChart, IDriveRemoteChart


class ActiveUsersChart:
    __slots__ = ('__data',)

    def __init__(self, data: IActiveUsersChart):
        self.__data = data

    @property
    def read_write(self) -> list[int]:
        return self.__data['read_write']

    @property
    def read(self) -> list[int]:
        return self.__data['read']

    @property
    def write(self) -> list[int]:
        return self.__data['write']

    @property
    def registered_within_week(self) -> list[int]:
        return self.__data['registered_within_week']

    @property
    def registered_within_month(self) -> list[int]:
        return self.__data['registered_within_month']

    @property
    def registered_within_year(self) -> list[int]:
        return self.__data['registered_within_year']

    @property
    def registered_outside_week(self) -> list[int]:
        return self.__data['registered_outside_week']

    @property
    def registered_outside_month(self) -> list[int]:
        return self.__data['registered_outside_month']

    @property
    def registered_outside_year(self) -> list[int]:
        return self.__data['registered_outside_year']


class DriveLocalChart:
    __slots__ = ('__data',)

    def __init__(self, data: IDriveLocalChart):
        self.__data = data

    @property
    def total_count(self) -> list[int]:
        return self.__data['total_count']

    @property
    def total_size(self) -> list[int]:
        return self.__data['total_size']

    @property
    def inc_count(self) -> list[int]:
        return self.__data['inc_count']

    @property
    def inc_size(self) -> list[int]:
        return self.__data['inc_size']

    @property
    def dec_count(self) -> list[int]:
        return self.__data['dec_count']

    @property
    def dec_size(self) -> list[int]:
        return self.__data['dec_size']


class DriveRemoteChart:
    __slots__ = ('__data',)

    def __init__(self, data: IDriveRemoteChart):
        self.__data: IDriveRemoteChart = data

    @property
    def total_count(self) -> list[int]:
        return self.__data['total_count']

    @property
    def total_size(self) -> list[int]:
        return self.__data['total_size']

    @property
    def inc_count(self) -> list[int]:
        return self.__data['inc_count']

    @property
    def inc_size(self) -> list[int]:
        return self.__data['inc_size']

    @property
    def dec_count(self) -> list[int]:
        return self.__data['dec_count']

    @property
    def dec_size(self) -> list[int]:
        return self.__data['dec_size']


class DriveChart:
    __slots__ = ('__data',)

    def __init__(self, data: IDriveChart):
        self.__data: IDriveChart = data

    @property
    def local(self) -> DriveLocalChart:
        return DriveLocalChart(self.__data['local'])

    @property
    def remote(self) -> DriveRemoteChart:
        return DriveRemoteChart(self.__data['remote'])



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.models.lite.meta import LiteMeta, MetaCommon
from mipac.types.meta import IAdminMeta, IFeatures, IMeta, IPolicies

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class Policies:
    def __init__(self, policies: IPolicies) -> None:
        self.__policies: IPolicies = policies

    @property
    def gtl_available(self) -> bool:
        return self.__policies['gtl_available']

    @property
    def ltl_available(self) -> bool:
        return self.__policies['ltl_available']

    @property
    def can_public_note(self) -> bool:
        return self.__policies['can_public_note']

    @property
    def can_invite(self) -> bool:
        return self.__policies['can_invite']

    @property
    def can_manage_custom_emojis(self) -> bool:
        return self.__policies['can_manage_custom_emojis']

    @property
    def can_hide_ads(self) -> bool:
        return self.__policies['can_hide_ads']

    @property
    def drive_capacity_mb(self) -> int:
        return self.__policies['drive_capacity_mb']

    @property
    def pin_limit(self) -> int:
        return self.__policies['pin_limit']

    @property
    def antenna_limit(self) -> int:
        return self.__policies['antenna_limit']

    @property
    def word_mute_limit(self) -> int:
        return self.__policies['word_mute_limit']

    @property
    def webhook_limit(self) -> int:
        return self.__policies['webhook_limit']

    @property
    def clip_limit(self) -> int:
        return self.__policies['clip_limit']

    @property
    def note_each_clips_limit(self) -> int:
        return self.__policies['note_each_clips_limit']

    @property
    def user_list_limit(self) -> int:
        return self.__policies['user_list_limit']

    @property
    def user_each_user_lists_limit(self) -> int:
        return self.__policies['user_each_user_lists_limit']

    @property
    def rate_limit_factor(self) -> int:
        return self.__policies['rate_limit_factor']


class Features:
    def __init__(self, features: IFeatures) -> None:
        self.__features = features

    @property
    def registration(self) -> bool:
        return self.__features['registration']

    @property
    def elasticsearch(self) -> bool:
        return self.__features['elasticsearch']

    @property
    def hcaptcha(self) -> bool:
        return self.__features['hcaptcha']

    @property
    def recaptcha(self) -> bool:
        return self.__features['recaptcha']

    @property
    def object_storage(self) -> bool:
        return self.__features['object_storage']

    @property
    def twitter(self) -> bool:
        return self.__features['twitter']

    @property
    def github(self) -> bool:
        return self.__features['github']

    @property
    def discord(self) -> bool:
        return self.__features['discord']

    @property
    def service_worker(self) -> bool:
        return self.__features['service_worker']

    @property
    def miauth(self) -> bool:
        return self.__features.get('miauth', False)

    @property
    def email_required_for_signup(self) -> bool:
        return self.__features.get('email_required_for_signup', False)

    @property
    def global_time_line(self) -> bool:
        return self.__features.get('global_time_line', False)

    @property
    def local_time_line(self) -> bool:
        return self.__features.get('local_time_line', False)


class Meta(LiteMeta):
    def __init__(self, meta: IMeta, *, client: ClientManager) -> None:
        super().__init__(meta, client=client)
        self.__meta = meta

    # shared

    @property
    def features(self) -> Features:
        return Features(self.__meta['features'])

    # v12

    @property
    def pinned_clip_id(self) -> str | None:
        return self.__meta.get('pinned_clip_id')

    @property
    def pinned_pages(self) -> list[str]:
        return self.__meta.get('pinned_pages', [])

    @property
    def policies(self) -> Policies | None:
        return Policies(self.__meta['policies']) if 'policies' in self.__meta else None

    @property
    def require_setup(self) -> bool:
        return self.__meta.get('require_setup', False)


class AdminMeta(MetaCommon):
    def __init__(self, admin_meta: IAdminMeta, *, client: ClientManager) -> None:
        super().__init__(admin_meta, client=client)
        self.__admin_meta = admin_meta

    # 全てのバージョン共通に存在する

    @property
    def drive_capacity_per_local_user_mb(self) -> int | None:
        return self.__admin_meta.get('drive_capacity_per_local_user_mb')

    @property
    def drive_capacity_per_remote_user_mb(self) -> int | None:
        return self.__admin_meta.get('drive_capacity_per_remote_user_mb')

    @property
    def hidden_tags(self) -> list[str]:
        return self.__admin_meta.get('hidden_tags', [])

    @property
    def blocked_hosts(self) -> list[str]:
        return self.__admin_meta.get('blocked_hosts', [])

    @property
    def recaptcha_secret_key(self) -> str | None:
        return self.__admin_meta.get('recaptcha_secret_key')

    @property
    def twitter_consumer_key(self) -> str | None:
        return self.__admin_meta.get('twitter_consumer_key')

    @property
    def twitter_consumer_secret(self) -> str | None:
        return self.__admin_meta.get('twitter_consumer_secret')

    @property
    def github_client_id(self) -> str | None:
        return self.__admin_meta.get('github_client_id')

    @property
    def github_client_secret(self) -> str | None:
        return self.__admin_meta.get('github_client_secret')

    @property
    def discord_client_id(self) -> str | None:
        return self.__admin_meta.get('discord_client_id')

    @property
    def discord_client_secret(self) -> str | None:
        return self.__admin_meta.get('discord_client_secret')

    @property
    def email(self) -> str | None:
        return self.__admin_meta.get('email')

    @property
    def smtp_secure(self) -> bool:
        return self.__admin_meta.get('smtp_secure', False)

    @property
    def smtp_host(self) -> str | None:
        return self.__admin_meta.get('smtp_host')

    @property
    def smtp_port(self) -> int | None:
        return self.__admin_meta.get('smtp_port')

    @property
    def smtp_user(self) -> str | None:
        return self.__admin_meta.get('smtp_user')

    @property
    def smtp_pass(self) -> str | None:
        return self.__admin_meta.get('smtp_pass')

    @property
    def sw_private_key(self) -> str | None:
        return self.__admin_meta.get('sw_private_key')

    @property
    def use_object_storage(self) -> bool:
        return self.__admin_meta.get('use_object_storage', False)

    @property
    def object_storage_base_url(self) -> str | None:
        return self.__admin_meta.get('object_storage_base_url')

    @property
    def object_storage_bucket(self) -> str | None:
        return self.__admin_meta.get('object_storage_bucket')

    @property
    def object_storage_prefix(self) -> str | None:
        return self.__admin_meta.get('object_storage_prefix')

    @property
    def object_storage_endpoint(self) -> str | None:
        return self.__admin_meta.get('object_storage_endpoint')

    @property
    def object_storage_region(self) -> str | None:
        return self.__admin_meta.get('object_storage_region')

    @property
    def object_storage_port(self) -> int | None:
        return self.__admin_meta.get('object_storage_port')

    @property
    def object_storage_access_key(self) -> str | None:
        return self.__admin_meta.get('object_storage_access_key')

    @property
    def object_storage_secret_key(self) -> str | None:
        return self.__admin_meta.get('object_storage_secret_key')

    @property
    def object_storage_use_ssl(self) -> bool:
        return self.__admin_meta.get('object_storage_use_ssl', False)

    @property
    def object_storage_use_proxy(self) -> bool:
        return self.__admin_meta.get('object_storage_use_proxy', False)

    @property
    def object_storage_set_public_read(self) -> bool:
        return self.__admin_meta.get('object_storage_set_public_read', False)

    @property
    def pinned_users(self) -> list[str]:
        return self.__admin_meta.get('pinned_users', [])

    """
    v12 only
    """

    @property
    def hcaptcha_secret_key(self) -> str | None:
        return self.__admin_meta.get('hcaptcha_secret_key')

    @property
    def sensitive_media_detection(self) -> str | None:
        return self.__admin_meta.get('sensitive_media_detection')

    @property
    def sensitive_media_detection_sensitivity(self) -> str | None:
        return self.__admin_meta.get('sensitive_media_detection_sensitivity')

    @property
    def set_sensitive_flag_automatically(self) -> bool:
        return self.__admin_meta.get('set_sensitive_flag_automatically', False)

    @property
    def enable_sensitive_media_detection_for_videos(self) -> bool:
        return self.__admin_meta.get('enable_sensitive_media_detection_for_videos', False)

    @property
    def proxy_account_id(self) -> str | None:
        return self.__admin_meta.get('proxy_account_id')

    @property
    def summary_proxy(self) -> str | None:
        return self.__admin_meta.get('summary_proxy')

    @property
    def enable_ip_logging(self) -> bool:
        return self.__admin_meta.get('enable_ip_logging', False)

    @property
    def enable_active_email_validation(self) -> bool:
        return self.__admin_meta.get('enable_active_email_validation', False)



from typing import Literal

IT_ACHIEVEMENT_NAME = Literal[
    'notes1',
    'notes10',
    'notes100',
    'notes500',
    'notes1000',
    'notes5000',
    'notes10000',
    'notes20000',
    'notes30000',
    'notes40000',
    'notes50000',
    'notes60000',
    'notes70000',
    'notes80000',
    'notes90000',
    'notes100000',
    'login3',
    'login7',
    'login15',
    'login30',
    'login60',
    'login100',
    'login200',
    'login300',
    'login400',
    'login500',
    'login600',
    'login700',
    'login800',
    'login900',
    'login1000',
    'passedSinceAccountCreated1',
    'passedSinceAccountCreated2',
    'passedSinceAccountCreated3',
    'loggedInOnBirthday',
    'loggedInOnNewYearsDay',
    'noteClipped1',
    'noteFavorited1',
    'myNoteFavorited1',
    'profileFilled',
    'markedAsCat',
    'following1',
    'following10',
    'following50',
    'following100',
    'following300',
    'followers1',
    'followers10',
    'followers50',
    'followers100',
    'followers300',
    'followers500',
    'followers1000',
    'collectAchievements30',
    'viewAchievements3min',
    'iLoveMisskey',
    'foundTreasure',
    'client30min',
    'noteDeletedWithin1min',
    'postedAtLateNight',
    'postedAt0min0sec',
    'selfQuote',
    'htl20npm',
    'viewInstanceChart',
    'outputHelloWorldOnScratchpad',
    'open3windows',
    'driveFolderCircularReference',
    'reactWithoutRead',
    'clickedClickHere',
    'justPlainLucky',
    'setNameToSyuilo',
    'cookieClicked',
    'brainDiver',
]



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.abstract.manager import AbstractManager
from mipac.http import HTTPClient, Route

if TYPE_CHECKING:
    from mipac.client import ClientManager


class PagesManager(AbstractManager):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self.__session: HTTPClient = session
        self.__client: ClientManager = client

    async def get_pages(
        self, limit: int = 100, since_id: int | None = None, until_id: int | None = None,
    ):
        data = {'limit': limit, 'since_id': since_id, 'until_id': until_id}
        res = await self.__session.request(Route('POST', '/api/i/pages'), json=data, auth=True)
        return res



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Literal

from mipac.types.ads import IAd
from mipac.util import str_to_datetime

if TYPE_CHECKING:
    from mipac.manager.admins.ad import AdminAdvertisingModelManager
    from mipac.manager.client import ClientManager


class Ad:
    def __init__(self, ad_data: IAd, *, client: ClientManager) -> None:
        self.__ad_data: IAd = ad_data
        self.__client: ClientManager = client

    @property
    def id(self) -> str:
        return self.__ad_data['id']

    @property
    def created_at(self) -> datetime:
        return str_to_datetime(self.__ad_data['created_at'])

    @property
    def starts_at(self) -> datetime:
        return str_to_datetime(self.__ad_data['id'])

    @property
    def expires_at(self) -> datetime:
        return str_to_datetime(self.__ad_data['id'])

    @property
    def url(self) -> str:
        return self.__ad_data['url']

    @property
    def place(self) -> Literal['square' 'horizontal' 'horizontal-big']:
        return self.__ad_data['place']

    @property
    def priority(self) -> Literal['high' 'middle' 'low']:
        return self.__ad_data['priority']

    @property
    def ratio(self) -> int:
        return self.__ad_data['ratio']

    @property
    def image_url(self) -> str:
        return self.__ad_data['image_url']

    @property
    def memo(self) -> str | None:
        return self.__ad_data['memo']

    @property
    def api(self) -> AdminAdvertisingModelManager:
        return self.__client.admin.create_ad_model_manager(ad_id=self.id)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from mipac.models.lite.user import LiteUser
from mipac.types.roles import IRole, IRolePolicies, IRolePolicieValue, IRoleUser
from mipac.util import str_to_datetime

if TYPE_CHECKING:
    from mipac.manager.admins.roles import AdminRolesModelManager
    from mipac.manager.client import ClientManager
    from mipac.manager.user import UserManager


class RoleUser:
    def __init__(self, role_user: IRoleUser, *, client: ClientManager) -> None:
        self.__role_user = role_user
        self.__client = client

    @property
    def id(self) -> str:
        return self.__role_user['id']

    @property
    def user(self) -> LiteUser:
        return LiteUser(self.__role_user['user'], client=self.__client)

    @property
    def expires_at(self) -> datetime | None:
        return (
            str_to_datetime(self.__role_user['expires_at'])
            if self.__role_user['expires_at']
            else None
        )

    @property
    def action(self) -> UserManager:
        return self.__client._create_user_instance(self.user)


class RolePolicyValue:
    def __init__(self, policiy_value_data: IRolePolicieValue) -> None:
        self.__policy_value_data = policiy_value_data

    @property
    def value(self) -> int:
        return self.__policy_value_data.get('value')

    @property
    def use_default(self) -> bool:
        return self.__policy_value_data.get('use_default')

    @property
    def priority(self) -> int | None:
        return self.__policy_value_data.get('priority')


class RolePolicies:
    def __init__(self, role_policies_data: IRolePolicies) -> None:
        self.__role_policies_data = role_policies_data

    @property
    def antenna_limit(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('antenna_limit'))

    @property
    def gtl_available(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('gtl_available'))

    @property
    def ltl_available(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('ltl_available'))

    @property
    def can_public_note(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('can_public_note'))

    @property
    def drive_capacity_mb(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('drive_capacity_mb'))

    @property
    def can_invite(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('can_invite'))

    @property
    def can_manage_custom_emojis(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('can_manage_custom_emojis'))

    @property
    def can_hide_ads(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('can_hide_ads'))

    @property
    def pin_limit(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('pin_limit'))

    @property
    def word_mute_limit(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('word_mute_limit'))

    @property
    def webhook_limit(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('webhook_limit'))

    @property
    def clip_limit(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('clip_limit'))

    @property
    def note_each_clips_limit(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('note_each_clips_limit'))

    @property
    def user_list_limit(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('user_list_limit'))

    @property
    def user_each_user_lists_limit(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('user_each_user_lists_limit'))

    @property
    def rate_limit_factor(self) -> RolePolicyValue:
        return RolePolicyValue(self.__role_policies_data.get('rate_limit_factor'))


class Role:
    def __init__(self, role_data: IRole, *, client: ClientManager) -> None:
        self.__role_data: IRole = role_data
        self.__client: ClientManager = client

    @property
    def id(self) -> str:
        return self.__role_data.get('id')

    @property
    def created_at(self) -> str:
        return self.__role_data.get('created_at')

    @property
    def updated_at(self) -> str:
        return self.__role_data.get('updated_at')

    @property
    def name(self) -> str:
        return self.__role_data.get('name')

    @property
    def description(self) -> str:
        return self.__role_data.get('description')

    @property
    def color(self) -> str | None:
        return self.__role_data.get('color')

    @property
    def icon_url(self) -> str | None:
        return self.__role_data.get('icon_url')

    @property
    def target(self) -> str:
        return self.__role_data.get('target')

    @property
    def cond_formula(self) -> dict:
        return self.__role_data.get('cond_formula')

    @property
    def is_public(self) -> bool:
        return self.__role_data.get('is_public')

    @property
    def is_administrator(self) -> bool:
        return self.__role_data.get('is_administrator')

    @property
    def is_moderator(self) -> bool:
        return self.__role_data.get('is_moderator')

    @property
    def as_badge(self) -> bool:
        return self.__role_data.get('as_badge')

    @property
    def can_edit_members_by_moderator(self) -> bool:
        return self.__role_data.get('can_edit_members_by_moderator')

    @property
    def policies(self) -> RolePolicies:
        return RolePolicies(self.__role_data.get('policies'))

    @property
    def users_count(self) -> int:
        return self.__role_data.get('users_count')

    @property
    def api(self) -> AdminRolesModelManager:
        return self.__client.admin.create_roles_model_manager(self.id)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from mipac.models.lite.instance import LiteInstance
from mipac.types.emoji import ICustomEmojiLite
from mipac.types.user import ILiteUser
from mipac.util import deprecated

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager
    from mipac.manager.user import UserManager


class LiteUser:
    __slots__ = ('__user', '__client')

    def __init__(self, user: ILiteUser, *, client: ClientManager) -> None:
        self.__user: ILiteUser = user
        self.__client: ClientManager = client

    @property
    def id(self) -> str:
        return self.__user['id']

    @property
    def username(self) -> str:
        return self.__user['username']

    @property
    def host(self) -> str | None:
        return self.__user['host'] if 'host' in self.__user else None

    @property
    @deprecated
    def name(self) -> str:
        return self.__user['name']

    @property
    def nickname(self) -> str:
        return self.__user['name']

    @property
    def online_status(self,) -> Literal['online', 'active', 'offline', 'unknown']:
        return self.__user['online_status']

    @property
    def avatar_url(self) -> str:
        return self.__user['avatar_url']

    @property
    def avatar_blurhash(self) -> str:
        return self.__user['avatar_blurhash']

    @property
    def avatar_color(self) -> str | None:
        """
        Returns the average color of the avatar.
        Note: Since avatar_color is deprecated in v13,
        only None is returned for v13 instances.

        Returns
        -------
        str | None
            average color of the avatar
        """

        return self.__user.get('avatar_color')

    @property
    def emojis(self) -> list[ICustomEmojiLite]:  # TODO: ちゃんとモデルにする
        """
        List of emoji included in nicknames, etc
        Note: emojis have been abolished since misskey v13

        Returns
        -------
        list[ICustomEmojiLite]
            List of emoji included in nicknames, etc
        """

        return self.__user.get('emojis', [])

    @property
    def instance(self) -> LiteInstance | None:
        return LiteInstance(self.__user['instance']) if 'instance' in self.__user else None

    @property
    def api(self) -> UserManager:
        return self.__client._create_user_instance(self)



from importlib import import_module
from typing import Union


class APIError(Exception):
    """APIのエラー"""

    def __init__(self, data: Union[dict, str], status: int):
        self.raw: Union[dict, str] = data
        self.status: int = status
        self.code: str | None = None
        self.id: str | None = None
        self.message: str | None = None
        if isinstance(data, dict):
            error = data.get('error', {})
            if isinstance(error, dict):
                self.code: str | None = error.get('code', '')
                self.id: str | None = error.get('id')
                self.message: str | None = error.get('message', '')
        super().__init__(self.message or self.raw)

    def raise_error(self):
        if not self.code:
            raise self
        if value := getattr(
            import_module('mipac.errors.errors'),
            ''.join([i.capitalize() for i in self.code.split('_')]) + 'Error',
            None,
        ):
            raise value(self.raw, self.status)
        else:
            raise self


class NotExistRequiredData(Exception):
    """クラスの中に必要なデータが不足している"""


class ParameterError(Exception):
    """引数に関するエラー"""


class NotSupportVersion(Exception):
    """サポートされていないバージョンのインスタンス"""


NotSupportVersionText = 'ご利用のインスタンスのバージョンではサポートされていない機能です'



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, Literal, overload

from mipac.abstract.action import AbstractAction
from mipac.errors.base import ParameterError
from mipac.http import HTTPClient, Route
from mipac.models.announcement import Announcement
from mipac.models.lite.meta import LiteMeta
from mipac.models.meta import Meta
from mipac.types.announcement import IAnnouncement
from mipac.types.meta import ILiteMeta, IMeta

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class ClientActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager) -> None:
        self.__session: HTTPClient = session
        self.__client: ClientManager = client

    @overload
    async def get_meta(self, detail: Literal[False] = ...) -> LiteMeta:
        ...

    @overload
    async def get_meta(self, detail: Literal[True] = ...) -> Meta:
        ...

    async def get_meta(self, detail: bool = False):
        params = {
            'route': Route('POST', '/api/meta'),
            'json': {'detail': detail},
            'auth': True,
            'lower': True,
        }
        if detail is True:
            meta: IMeta = await self.__session.request(**params)
            return Meta(meta, client=self.__client)
        lite_meta: ILiteMeta = await self.__session.request(**params)
        return LiteMeta(lite_meta, client=self.__client)

    async def get_announcements(
        self,
        limit: int = 10,
        with_unreads: bool = False,
        since_id: str | None = None,
        until_id: str | None = None,
        *,
        all: bool = False
    ) -> AsyncGenerator[Announcement, None]:
        if limit > 100:
            raise ParameterError('limitは100以下である必要があります')
        if all:
            limit = 100

        async def request(req_body) -> list[Announcement]:
            res: list[IAnnouncement] = await self.__session.request(
                Route('POST', '/api/announcements'), auth=True, json=req_body,
            )
            return [Announcement(announcement, client=self.__client) for announcement in res]

        body = {
            'limit': limit,
            'withUnreads': with_unreads,
            'sinceId': since_id,
            'untilId': until_id,
        }
        first_req = await request(body)

        for announcement in first_req:
            yield announcement
        if all and len(first_req) == 100:
            body['untilId'] = first_req[-1].id
            while True:
                res = await request(body)
                if len(res) <= 100:
                    for announcement in res:
                        yield announcement
                if len(res) < 100:
                    break



from typing import Any, Generic, Literal, NotRequired, Optional, TypedDict, TypeVar

from .drive import IDriveFile
from .emoji import ICustomEmojiLite
from .poll import IPoll
from .user import ILiteUser

T = TypeVar('T')


class INoteState(TypedDict):
    is_favorited: bool
    is_watching: bool
    is_muted_thread: NotRequired[bool]


class INoteUpdated(TypedDict, Generic[T]):
    type: Literal['noteUpdated']
    body: T


class INoteUpdatedDeleteBody(TypedDict):
    deleted_at: str


class INoteUpdatedDelete(TypedDict):
    id: str
    type: Literal['deleted']
    body: INoteUpdatedDeleteBody


class INoteUpdatedReactionBody(TypedDict):
    reaction: str
    emoji: ICustomEmojiLite
    user_id: str


class INoteUpdatedReaction(TypedDict):
    id: str
    type: Literal['reacted', 'unreacted']
    body: INoteUpdatedReactionBody


class GeoPayload(TypedDict):
    """
    衛星情報
    """

    coordinates: Optional[list[Any]]
    altitude: int | None
    accuracy: int | None
    altitude_accuracy: int | None
    heading: int | None
    speed: int | None


class INoteRequired(TypedDict):
    id: str
    created_at: str
    text: str | None
    cw: str | None
    user: ILiteUser
    user_id: str
    reply_id: str
    renote_id: str
    files: list[IDriveFile]
    file_ids: list[str]
    visibility: Literal['public', 'home', 'followers', 'specified']
    reactions: dict[str, int]
    renote_count: int
    replies_count: int


class INote(INoteRequired, total=False):
    """
    note object
    """

    renote: 'INote'
    reply: 'INote'
    visible_user_ids: list[str]
    local_only: bool
    my_reaction: str
    uri: str
    url: str
    is_hidden: bool
    poll: IPoll
    emojis: list[ICustomEmojiLite]
    reaction_emojis: dict[str, str]
    reaction_acceptance: Literal['likeOnly', 'likeOnlyForRemote'] | None


class ICreatedNote(TypedDict):
    """
    created note
    """

    created_note: INote


class INoteReaction(TypedDict):
    id: str
    created_at: str
    user: ILiteUser
    type: str


class INoteTranslateResult(TypedDict):
    sourceLang: str
    text: str



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator

from mipac.abstract.action import AbstractAction
from mipac.errors.base import ParameterError
from mipac.http import HTTPClient, Route
from mipac.models.announcement import Announcement, AnnouncementSystem
from mipac.types.announcement import IAnnouncement, IAnnouncementSystem

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class AdminAnnouncementClientActions(AbstractAction):
    def __init__(
        self, announce_id: str | None = None, *, session: HTTPClient, client: ClientManager,
    ):
        self.__announce_id = announce_id
        self.__session: HTTPClient = session
        self.__client: ClientManager = client

    async def delete(self, announce_id: str | None = None) -> bool:
        announce_id = announce_id or self.__announce_id
        res: bool = await self.__session.request(
            Route('POST', '/api/admin/announcements/delete'), json={'id': announce_id}, auth=True,
        )
        return res

    async def update(
        self,
        title: str,
        text: str,
        image_url: str | None = None,
        *,
        announce_id: str | None = None,
    ):
        announce_id = announce_id or self.__announce_id
        body = {
            'id': announce_id,
            'title': title,
            'text': text,
            'imageUrl': image_url,
        }
        res: bool = await self.__session.request(
            Route('POST', '/api/admin/announcements/update'),
            json=body,
            auth=True,
            remove_none=False,
        )
        return res


class AdminAnnouncementActions(AdminAnnouncementClientActions):
    def __init__(
        self, announce_id: str | None = None, *, session: HTTPClient, client: ClientManager,
    ):
        super().__init__(announce_id=announce_id, session=session, client=client)

    async def create(self, title: str, text: str, image_url: str | None = None) -> Announcement:
        body = {'title': title, 'text': text, 'imageUrl': image_url}
        created_announcement: IAnnouncement = await self.__session.request(
            Route('POST', '/api/admin/announcements/create'),
            json=body,
            auth=True,
            lower=True,
            remove_none=False,
        )
        return Announcement(created_announcement, client=self.__client)

    async def gets(
        self, limit: int = 10, since_id: str | None = None, until_id: str | None = None,
    ) -> AsyncGenerator[AnnouncementSystem, None]:
        if limit > 100:
            raise ParameterError('limitは100以下である必要があります')
        if all:
            limit = 100

        async def request(req_body) -> list[AnnouncementSystem]:
            res: list[IAnnouncementSystem] = await self.__session.request(
                Route('POST', '/api/admin/announcements/list'), auth=True, json=req_body,
            )
            return [AnnouncementSystem(announcement, client=self.__client) for announcement in res]

        body = {
            'limit': limit,
            'sinceId': since_id,
            'untilId': until_id,
        }
        first_req = await request(body)

        for announcement in first_req:
            yield announcement
        if all and len(first_req) == 100:
            body['untilId'] = first_req[-1].id
            while True:
                res = await request(body)
                if len(res) <= 100:
                    for announcement in res:
                        yield announcement
                if len(res) < 100:
                    break



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.models.user import UserDetailed
from mipac.types.mute import IMuteUser

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class MuteUser:
    def __init__(self, data: IMuteUser, *, client: ClientManager) -> None:
        self.__data: IMuteUser = data
        self.__client: ClientManager = client

    @property
    def id(self) -> str:
        return self.__data['id']

    @property
    def created_at(self) -> str:
        return self.__data['created_at']

    @property
    def mutee_id(self) -> str:
        return self.__data['mutee_id']

    @property
    def mutee(self) -> UserDetailed:
        return UserDetailed(self.__data['mutee'], client=self.__client)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Generic, TypeVar

from mipac.types.channel import IChannelLite
from mipac.util import str_to_datetime

if TYPE_CHECKING:
    from mipac.manager import ClientManager

T = TypeVar('T', bound=IChannelLite)


class ChannelLite(Generic[T]):
    def __init__(self, channel: T, *, client: ClientManager) -> None:
        self._channel: T = channel
        self.__client: ClientManager = client

    @property
    def id(self) -> str:
        return self._channel['id']

    @property
    def created_at(self) -> datetime:
        return str_to_datetime(self._channel['created_at'])

    @property
    def last_noted_at(self) -> datetime | None:
        last_noted_at = self._channel.get('last_noted_at')
        return str_to_datetime(last_noted_at) if last_noted_at else None

    @property
    def name(self) -> str:
        return self._channel['name']

    @property
    def description(self) -> str | None:
        return self._channel['description']

    @property
    def banner_url(self) -> str | None:
        return self._channel['banner_url']

    @property
    def notes_count(self) -> int:
        return self._channel['notes_count']

    @property
    def users_count(self) -> int:
        return self._channel['users_count']

    @property
    def is_following(self) -> bool:
        return bool(self._channel['is_following'])

    @property
    def user_id(self) -> str:
        return self._channel['user_id']



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.types.instance import IFederationInstance

if TYPE_CHECKING:
    from mipac.client import ClientManager


class FederationInstance:
    def __init__(self, instance: IFederationInstance, *, client: ClientManager) -> None:
        self.__instance: IFederationInstance = instance
        self.__client: ClientManager = client

    @property
    def id(self) -> str:
        return self.__instance['id']

    @property
    def host(self) -> str:
        return self.__instance['host']

    @property
    def users_count(self) -> int:
        return self.__instance['users_count']

    @property
    def notes_count(self) -> int:
        return self.__instance['notes_count']

    @property
    def following_count(self) -> int:
        return self.__instance['following_count']

    @property
    def followers_count(self) -> int:
        return self.__instance['followers_count']

    @property
    def is_not_responding(self) -> bool:
        return self.__instance['is_not_responding']

    @property
    def is_suspended(self) -> bool:
        return self.__instance['is_suspended']

    @property
    def is_blocked(self) -> bool:
        return self.__instance['is_blocked']

    @property
    def software_name(self) -> str:
        return self.__instance['software_name']

    @property
    def software_version(self) -> str:
        return self.__instance['software_version']

    @property
    def open_registrations(self) -> bool:
        return self.__instance['open_registrations']

    @property
    def name(self) -> str:
        return self.__instance['name']

    @property
    def description(self) -> str:
        return self.__instance['description']

    @property
    def maintainer_name(self) -> str:
        return self.__instance['maintainer_name']

    @property
    def maintainer_email(self) -> str:
        return self.__instance['maintainer_email']

    @property
    def icon_url(self) -> str:
        return self.__instance['icon_url']

    @property
    def favicon_url(self) -> str:
        return self.__instance['favicon_url']

    @property
    def theme_color(self) -> str:
        return self.__instance['theme_color']

    @property
    def info_updated_at(self) -> str:
        return self.__instance['info_updated_at']

    @property
    def caught_at(self) -> str | None:
        return self.__instance.get('caught_at')

    @property
    def first_retrieved_at(self) -> str | None:
        return self.__instance.get('first_retrieved_at')

    @property
    def latest_request_sent_at(self) -> str | None:
        return self.__instance.get('latest_request_sent_at')

    @property
    def last_communicated_at(self) -> str | None:
        return self.__instance.get('last_communicated_at')



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator

from mipac.abstract.action import AbstractAction
from mipac.http import HTTPClient, Route
from mipac.models.user import BlockingUser, UserDetailed
from mipac.types.user import IBlockingUser, IUserDetailed

if TYPE_CHECKING:
    from mipac.client import ClientManager


class BlockingActions(AbstractAction):
    def __init__(self, user_id: str | None = None, *, session: HTTPClient, client: ClientManager):
        self.__user_id: str | None = user_id
        self.__session: HTTPClient = session
        self.__client: ClientManager = client

    async def add(self, user_id: str | None = None) -> UserDetailed:
        user_id = self.__user_id or user_id
        res: IUserDetailed = await self.__session.request(
            Route('POST', '/api/blocking/create'), auth=True, json={'userId': user_id}, lower=True
        )
        return UserDetailed(res, client=self.__client)

    async def remove(self, user_id: str | None = None) -> UserDetailed:
        user_id = self.__user_id or user_id
        res: IUserDetailed = await self.__session.request(
            Route('POST', '/api/blocking/delete'), auth=True, json={'userId': user_id}, lower=True
        )
        return UserDetailed(res, client=self.__client)

    async def get_list(
        self,
        since_id: str | None = None,
        until_id: str | None = None,
        limit: int = 100,
        all: bool = False,
    ) -> AsyncGenerator[BlockingUser, None]:
        async def request(body) -> list[BlockingUser]:
            res: list[IBlockingUser] = await self.__session.request(
                Route('POST', '/api/blocking/list'), lower=True, auth=True, json=body
            )
            return [BlockingUser(user, client=self.__client) for user in res]

        data = {'limit': limit, 'sinceId': since_id, 'untilId': until_id}
        if all:
            data['limit'] = 100
        first_req = await request(data)
        for user in first_req:
            yield user

        if all and len(first_req) == 100:
            data['untilId'] = first_req[-1].id
            while True:
                res = await request(data)
                if len(res) <= 100:
                    for user in res:
                        yield user
                if len(res) == 0:
                    break
                data['untilId'] = res[-1].id



from __future__ import annotations

import re
import sys
from typing import Any, Literal, TypeVar

import aiohttp

from mipac import __version__
from mipac.config import config
from mipac.errors.base import APIError
from mipac.types.endpoints import ENDPOINTS
from mipac.types.meta import IMeta
from mipac.types.user import IUserDetailed
from mipac.util import _from_json, remove_dict_empty, upper_to_lower


class _MissingSentinel:
    def __eq__(self, other):
        return False

    def __bool__(self):
        return False

    def __repr__(self):
        return '...'


MISSING: Any = _MissingSentinel()
R = TypeVar('R')


class MisskeyClientWebSocketResponse(aiohttp.ClientWebSocketResponse):
    async def close(self, *, code: int = 4000, message: bytes = b'') -> bool:
        return await super().close(code=code, message=message)


async def json_or_text(response: aiohttp.ClientResponse):
    text = await response.text(encoding='utf-8')
    try:
        if 'application/json' in response.headers['Content-Type']:
            return _from_json(text)
    except KeyError:
        pass


class Route:
    def __init__(self, method: Literal['GET', 'POST'], path: ENDPOINTS):
        self.path: str = path
        self.method: str = method


class HTTPClient:
    def __init__(self, url: str, token: str | None = None) -> None:
        user_agent = (
            'Misskey Bot (https://github.com/yupix/MiPA {0})' + 'Python/{1[0]}.{1[1]} aiohttp/{2}'
        )
        self.user_agent = user_agent.format(__version__, sys.version_info, aiohttp.__version__)
        self._session: aiohttp.ClientSession = MISSING
        self._url: str = url
        self._token: str | None = token

    @property
    def session(self) -> aiohttp.ClientSession:
        return self._session

    async def request(
        self, route: Route, auth: bool = False, remove_none: bool = True, **kwargs
    ) -> R:
        headers: dict[str, str] = {
            'User-Agent': self.user_agent,
        }

        is_lower = kwargs.pop('lower') if kwargs.get('lower') else False

        if 'json' in kwargs:
            headers['Content-Type'] = 'application/json'
            kwargs['json'] = kwargs.pop('json')

        if auth:
            key = 'json' if 'json' in kwargs or 'data' not in kwargs else 'data'
            if not kwargs.get(key):
                kwargs[key] = {}
            if self._token:
                kwargs[key]['i'] = self._token

        replace_list = kwargs.pop('replace_list', {})

        for i in ('json', 'data'):
            if kwargs.get(i) and remove_none:
                kwargs[i] = remove_dict_empty(kwargs[i])
        async with self._session.request(route.method, self._url + route.path, **kwargs) as res:
            data = await json_or_text(res)
            if is_lower:
                if isinstance(data, list):
                    data = [upper_to_lower(i, replace_list=replace_list) for i in data]
                if isinstance(data, dict):
                    data = upper_to_lower(data)
            if res.status == 204 and data is None:
                return True  # type: ignore
            if 300 > res.status >= 200:
                return data  # type: ignore
            if 511 > res.status >= 300:
                if isinstance(data, dict):
                    APIError(data, res.status).raise_error()
            APIError('HTTP ERROR', res.status).raise_error()

    async def close_session(self) -> None:
        await self._session.close()

    async def login(self) -> IUserDetailed | None:
        match_domain = re.search(r'https?:\/\/([^\/]+)', self._url)
        match_protocol = re.search(r'^(http|https)', self._url)
        if match_domain is None or match_protocol is None:
            raise Exception()
        protocol = True if match_protocol.group(1) == 'https' else False
        config.from_dict(
            host=match_domain.group(1), is_ssl=protocol,
        )
        self._session = aiohttp.ClientSession(ws_response_class=MisskeyClientWebSocketResponse)
        if self._token:
            data: IUserDetailed = await self.request(Route('POST', '/api/i'), auth=True)
            if config.use_version_autodetect:
                meta: IMeta = await self.request(Route('POST', '/api/meta'), auth=True)
                use_version = int(meta['version'].split('.')[0])
                if isinstance(use_version, int) and use_version in (13, 12, 11):
                    config.use_version = use_version
            return data



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator

from mipac.abstract.action import AbstractAction
from mipac.errors.base import ParameterError
from mipac.http import HTTPClient, Route
from mipac.models.mute import MuteUser
from mipac.types.mute import IMuteUser
from mipac.util import remove_dict_empty

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class MuteActions(AbstractAction):
    def __init__(self, user_id: str | None = None, *, session: HTTPClient, client: ClientManager):
        self._user_id: str | None = user_id
        self._session: HTTPClient = session
        self._client: ClientManager = client

    async def add(self, user_id: str | None = None) -> bool:
        """
        Adds the specified user as a mute target

        Parameters
        ----------
        user_id : str | None, optional
            Mute target user Id, by default None

        Returns
        -------
        bool
            Whether the mute was successful or not
        """

        user_id = user_id or self._user_id
        data = {'userId': user_id}
        res: bool = await self._session.request(
            Route('POST', '/api/mute/create'), auth=True, json=data
        )
        return res

    async def delete(self, user_id: str | None = None) -> bool:
        """
        Unmute the specified user

        Parameters
        ----------
        user_id : str | None, optional
            Unmute target user Id, by default None

        Returns
        -------
        bool
            Whether the unmute was successful or not.
        """

        user_id = user_id or self._user_id
        data = {'userId': user_id}
        res: bool = await self._session.request(
            Route('POST', '/api/mute/delete'), auth=True, json=data
        )
        return res

    async def gets(
        self,
        limit: int = 100,
        since_id: str | None = None,
        until_id: str | None = None,
        all: bool = True,
    ) -> AsyncGenerator[MuteUser, None]:
        if limit > 100:
            raise ParameterError('limit は100以下である必要があります')

        async def request(body) -> list[MuteUser]:
            res: list[IMuteUser] = await self._session.request(
                Route('POST', '/api/mute/list'), lower=True, auth=True, json=body,
            )
            return [MuteUser(user, client=self._client) for user in res]

        data = remove_dict_empty({'limit': limit, 'sinceId': since_id, 'untilId': until_id})

        if all:
            data['limit'] = 100
        first_req = await request(data)

        for user in first_req:
            yield user

        if all and len(first_req) == 100:
            data['untilId'] = first_req[-1].id
            while True:
                res = await request(data)
                if len(res) <= 100:
                    for user in res:
                        yield user
                if len(res) == 0:
                    break
                data['untilId'] = res[-1].id



from __future__ import annotations

from typing import TYPE_CHECKING

from mipac.abstract.action import AbstractAction
from mipac.errors.base import ParameterError
from mipac.http import HTTPClient, Route
from mipac.models.note import Note
from mipac.types.note import INote

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class PollActions(AbstractAction):
    def __init__(self, note_id: str | None = None, *, session: HTTPClient, client: ClientManager):
        self.__note_id: str | None = note_id
        self.__session: HTTPClient = session
        self.__client: ClientManager = client

    async def vote(self, choice: int, note_id: str | None = None) -> bool:
        note_id = note_id or self.__note_id

        if note_id is None:
            raise ParameterError('note_id is required')

        data = {'noteId': note_id, 'choice': choice}
        res: bool = await self.__session.request(
            Route('POST', '/api/notes/polls/vote'), auth=True, json=data
        )
        return res

    async def recommendation(self, limit: int = 100, offset: int = 0, all: bool = True):

        if limit > 100:
            raise ParameterError('limit must be less than 100')

        if all:
            limit = 100

        async def request(body) -> list[Note]:
            res: list[INote] = await self.__session.request(
                Route('POST', '/api/notes/polls/recommendation'), lower=True, auth=True, json=body,
            )
            return [Note(note, client=self.__client) for note in res]

        data = {'limit': limit, 'offset': offset}
        first_res: list[Note] = await request(data)
        for note in first_res:
            yield note
        offset = limit
        count = 1
        if all and len(first_res) == limit:
            data['offset'] = limit * count
            while True:
                res = await request(data)
                if len(res) <= limit:
                    for note in res:
                        yield note
                if len(res) == 0 or len(res) < limit:
                    break
                data['offset'] = limit * count
                count = count + 1



from mipac.config import Config, config
from mipac.http import HTTPClient
from mipac.manager.client import ClientManager


class Client:
    def __init__(self, url: str, token: str | None = None) -> None:
        self.__url: str = url
        self.__token: str | None = token
        self.config: Config = config
        self.http: HTTPClient = HTTPClient(url, token)

    @property
    def api(self) -> ClientManager:
        return ClientManager(self.http, self.config)

    async def close_session(self) -> None:
        await self.http.close_session()



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, Literal, Optional

from mipac.config import config
from mipac.errors.base import NotExistRequiredData, NotSupportVersion, ParameterError
from mipac.http import HTTPClient, Route
from mipac.models.user import Achievement, LiteUser, UserDetailed
from mipac.util import cache, check_multi_arg, remove_dict_empty

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager
    from mipac.models.note import Note

__all__ = ['UserActions']


class UserActions:
    def __init__(
        self, session: HTTPClient, client: ClientManager, user: Optional[LiteUser] = None,
    ):
        self.__session: HTTPClient = session
        self.__user: Optional[LiteUser] = user
        self.__client: ClientManager = client

    async def get_me(self) -> UserDetailed:
        """
        ログインしているユーザーの情報を取得します
        """

        res = await self.__session.request(Route('POST', '/api/i'), auth=True, lower=True,)
        return UserDetailed(res, client=self.__client)  # TODO: 自分用のクラスに変更する

    def get_profile_link(
        self, external: bool = True, protocol: Literal['http', 'https'] = 'https',
    ):
        if not self.__user:
            return None
        host = (
            f'{protocol}://{self.__user.host}'
            if external and self.__user.host
            else self.__session._url
        )
        path = (
            f'/@{self.__user.username}' if external else f'/{self.__user.api.action.get_mention()}'
        )
        return host + path

    @cache(group='get_user')
    async def get(
        self,
        user_id: str | None = None,
        username: str | None = None,
        host: str | None = None,
        **kwargs,
    ) -> UserDetailed:
        """
        ユーザーのプロフィールを取得します。一度のみサーバーにアクセスしキャッシュをその後は使います。
        fetch_userを使った場合はキャッシュが廃棄され再度サーバーにアクセスします。

        Parameters
        ----------
        user_id : str
            取得したいユーザーのユーザーID
        username : str
            取得したいユーザーのユーザー名
        host : str, default=None
            取得したいユーザーがいるインスタンスのhost

        Returns
        -------
        UserDetailed
            ユーザー情報
        """

        field = remove_dict_empty({'userId': user_id, 'username': username, 'host': host})
        data = await self.__session.request(
            Route('POST', '/api/users/show'), json=field, auth=True, lower=True
        )
        return UserDetailed(data, client=self.__client)

    async def fetch(
        self, user_id: str | None = None, username: str | None = None, host: str | None = None,
    ) -> UserDetailed:
        """
        サーバーにアクセスし、ユーザーのプロフィールを取得します。基本的には get_userをお使いください。

        Parameters
        ----------
        user_id : str
            取得したいユーザーのユーザーID
        username : str
            取得したいユーザーのユーザー名
        host : str, default=None
            取得したいユーザーがいるインスタンスのhost

        Returns
        -------
        UserDetailed
            ユーザー情報
        """
        return await self.get(user_id=user_id, username=username, host=host, cache_override=True)

    async def get_notes(
        self,
        user_id: str | None = None,
        include_replies: bool = True,
        limit: int = 10,
        since_id: str | None = None,
        until_id: str | None = None,
        since_date: int = 0,
        until_date: int = 0,
        include_my_renotes: bool = True,
        with_files: bool = False,
        file_type: Optional[list[str]] = None,
        exclude_nsfw: bool = True,
    ) -> list[Note]:

        if check_multi_arg(user_id, self.__user):
            raise ParameterError('user_idがありません')

        user_id = user_id or self.__user and self.__user.id
        data = {
            'userId': user_id,
            'includeReplies': include_replies,
            'limit': limit,
            'sinceId': since_id,
            'untilId': until_id,
            'sinceDate': since_date,
            'untilDate': until_date,
            'includeMyRenotes': include_my_renotes,
            'withFiles': with_files,
            'fileType': file_type,
            'excludeNsfw': exclude_nsfw,
        }
        res = await self.__session.request(
            Route('POST', '/api/users/notes'), json=data, auth=True, lower=True
        )
        return [Note(i, client=self.__client) for i in res]

    def get_mention(self, user: Optional[LiteUser] = None) -> str:
        """
        Get mention name of user.

        Parameters
        ----------
        user : Optional[User], default=None
            メンションを取得したいユーザーのオブジェクト

        Returns
        -------
        str
            メンション
        """

        user = user or self.__user

        if user is None:
            raise NotExistRequiredData('Required parameters: user')
        return f'@{user.username}@{user.host}' if user.instance else f'@{user.username}'

    async def search(
        self,
        query: str,
        limit: int = 100,
        offset: int = 0,
        origin: Literal['local', 'remote', 'combined'] = 'combined',
        detail: bool = True,
        *,
        all: bool = False,
    ) -> AsyncGenerator[UserDetailed | LiteUser, None]:
        """
        Search users by keyword.

        Parameters
        ----------
        query : str
            Keyword to search.
        limit : int, default=100
            The maximum number of users to return.
        offset : int, default=0
            The number of users to skip.
        origin : Literal['local', 'remote', 'combined'], default='combined'
            The origin of users to search.
        detail : bool, default=True
            Whether to return detailed user information.
        all : bool, default=False
            Whether to return all users.

        Returns
        -------
        AsyncGenerator[UserDetailed | LiteUser, None]
            A AsyncGenerator of users.
        """

        if limit > 100:
            raise ParameterError('limit は100以下である必要があります')

        async def request(body) -> list[UserDetailed | LiteUser]:
            res = await self.__session.request(
                Route('POST', '/api/users/search'), lower=True, auth=True, json=body,
            )
            return [
                UserDetailed(user, client=self.__client)
                if detail
                else LiteUser(user, client=self.__client)
                for user in res
            ]

        body = remove_dict_empty(
            {'query': query, 'limit': limit, 'offset': offset, 'origin': origin, 'detail': detail}
        )

        if all:
            body['limit'] = 100
        first_req = await request(body)

        for user in first_req:
            yield user

        if all and len(first_req) == 100:
            times = 1
            while True:
                body['offset'] = times * 100
                res = await request(body)
                if len(res) <= 100:
                    for user in res:
                        yield user
                if len(res) == 0:
                    break
                times += 1

    async def search_by_username_and_host(
        self, username: str, host: str, limit: int = 100, detail: bool = True,
    ) -> list[UserDetailed | LiteUser]:
        """
        Search users by username and host.

        Parameters
        ----------
        username : str
            Username of user.
        host : str
            Host of user.
        limit : int, default=100
            The maximum number of users to return.
        detail : bool, default=True
            Weather to get detailed user information.

        Returns
        -------
        list[UserDetailed | LiteUser]
            A list of users.
        """

        if limit > 100:
            raise ParameterError('limit は100以下である必要があります')

        body = remove_dict_empty(
            {'username': username, 'host': host, 'limit': limit, 'detail': detail}
        )
        res = await self.__session.request(
            Route('POST', '/api/users/search-by-username-and-host'),
            lower=True,
            auth=True,
            json=body,
        )
        return [
            UserDetailed(user, client=self.__client)
            if detail
            else LiteUser(user, client=self.__client)
            for user in res
        ]

    async def get_achievements(self, user_id: str | None = None) -> list[Achievement]:
        """ Get achievements of user. """

        if config.use_version < 13:
            raise NotSupportVersion('ご利用のインスタンスのバージョンではサポートされていない機能です')

        user_id = user_id or self.__user and self.__user.id

        if not user_id:
            raise ParameterError('user_id is required')

        data = {
            'userId': user_id,
        }
        res = await self.__session.request(
            Route('POST', '/api/users/achievements'), json=data, auth=True, lower=True,
        )
        return [Achievement(i) for i in res]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from mipac.abstract.action import AbstractAction
from mipac.errors.base import ParameterError
from mipac.http import HTTPClient, Route
from mipac.models.instance import FederationInstance
from mipac.models.user import UserDetailed
from mipac.types.follow import IFederationFollower, IFederationFollowing
from mipac.types.instance import IFederationInstance, IFederationInstanceStat
from mipac.types.user import IUserDetailed

if TYPE_CHECKING:
    from mipac.manager.client import ClientManager


class FederationActions(AbstractAction):
    def __init__(self, *, session: HTTPClient, client: ClientManager):
        self.__session = session
        self.__client = client

    async def get_ap(self, uri: str) -> dict[Any, Any]:
        return dict(
            await self.__session.request(
                Route('POST', '/api/ap/get'), auth=True, json={'uri': uri}, lower=True
            )
        )

    async def show_ap(
        self, host: str, since_id: str | None = None, until_id: str | None = None, limit: int = 10
    ) -> FederationInstance:
        body = {'host': host, 'sinceId': since_id, 'untilId': until_id, 'limit': limit}

        res: FederationInstance = await self.__session.request(
            Route('POST', '/api/ap/show'), auth=True, json=body
        )
        return res

    async def get_followers(
        self, host: str, since_id: str | None = None, until_id: str | None = None, limit: int = 10
    ):
        body = {'host': host, 'sinceId': since_id, 'untilId': until_id, 'limit': limit}
        res: list[IFederationFollower] = await self.__session.request(
            Route('POST', '/api/federation/followers'), auth=True, json=body, lower=True
        )
        return res

    async def get_following(
        self, host: str, since_id: str | None = None, until_id: str | None = None, limit: int = 10
    ):
        body = {'host': host, 'sinceId': since_id, 'untilId': until_id, 'limit': limit}
        res: list[IFederationFollowing] = await self.__session.request(
            Route('POST', '/api/federation/following'), auth=True, json=body, lower=True
        )
        return res

    async def get_instances(
        self,
        host: str | None = None,
        blocked: bool | None = None,
        not_responding: bool | None = None,
        suspended: bool | None = None,
        federating: bool | None = None,
        subscribing: bool | None = None,
        publishing: bool | None = None,
        limit: int = 30,
        offset: int = 0,
        sort: str | None = None,
    ) -> list[FederationInstance]:
        if limit > 100:
            raise ParameterError('limitは100以下である必要があります')
        body = {
            'host': host,
            'blocked': blocked,
            'notResponding': not_responding,
            'suspended': suspended,
            'federating': federating,
            'subscribing': subscribing,
            'publishing': publishing,
            'limit': limit,
            'offset': offset,
            'sort': sort,
        }

        res: list[IFederationInstance] = await self.__session.request(
            Route('POST', '/api/federation/instances'), auth=True, lower=True, json=body
        )

        return [FederationInstance(i, client=self.__client) for i in res]

    async def show_instance(self, host: str) -> FederationInstance:
        res: IFederationInstance = await self.__session.request(
            Route('POST', '/api/federation/show-instance'),
            auth=True,
            json={'host': host},
            lower=True,
        )

        return FederationInstance(res, client=self.__client)

    async def update_remote_user(self, user_id: str) -> bool:
        return bool(
            self.__session.request(
                Route('POST', '/api/federation/update-remote-user'),
                auth=True,
                json={'userId': user_id},
            )
        )

    async def get_users(
        self, host: str, since_id: str | None = None, until_id: str | None = None, limit: int = 10
    ) -> UserDetailed:
        if limit > 100:
            raise ParameterError('limitは100以下である必要があります')
        body = {'host': host, 'sinceId': since_id, 'untilId': until_id, 'limit': limit}

        res: IUserDetailed = await self.__session.request(
            Route('POST', '/api/federation/users'), auth=True, json=body
        )
        return UserDetailed(res, client=self.__client)

    async def get_stats(self, limit: int = 10) -> IFederationInstanceStat:
        if limit > 100:
            raise ParameterError('limitは100以下である必要があります')
        res: IFederationInstanceStat = await self.__session.request(
            Route('POST', '/api/federation/stats'), auth=True, body={'limit': limit}, lower=True
        )
        return res
